/*
	pcx2g34.C

	Convert PCX images to G34 format.

	Version 1.0

	Copyright (C) 1994, Geoff Friesen B.Sc.
	All rights reserved.
*/

#if !defined(__LARGE__)
#error Large Memory Model Expected
#endif

#include <alloc.H>
#include <dir.H>
#include <dos.H>
#include <stdio.H>
#include <stdlib.H>
#include <string.H>

typedef struct
{
   char manufacturer;
   char version;
   char encoding;
   char bits_per_pixel;
   int xmin, ymin;
   int xmax, ymax;
   int hres;
   int vres;
   char palette [48];
   char reserved;
   char colour_planes;
   int bytes_per_line;
   int palette_type;
   char filler [58];
}
PCXHEAD;

int  Copy2		(FILE *fs, FILE *fd, int x, int y, int nx, int ny,
			 int bytes, char *path);
int  Copy8		(FILE *fs, FILE *fd, int x, int y, int nx, int ny,
			 int bytes, char *path);
int  Copy16		(FILE *fs, FILE *fd, int x, int y, int nx, int ny,
			 int bytes, char *path);
int  Copy256		(FILE *fs, FILE *fd, int x, int y, int nx, int ny,
			 int bytes, char *path);
int  ReadPcxLine	(char *p, FILE *fp, unsigned bytes);

int main (int argc, char **argv)
{
   FILE *fs, *fd;
   PCXHEAD pcxhdr;
   char drive [MAXDRIVE];
   char dir [MAXDIR];
   char file [MAXFILE];
   char ext [MAXEXT];
   char path [MAXPATH];
   char dac [768], header [128];
   unsigned bits, bytes, depth, i, nx, ny, status, width, x, y;

   fprintf (stderr, "ͻ\n");
   fprintf (stderr, " PCX2G34: Convert PCX images to G34      \n");
   fprintf (stderr, "                                         \n");
   fprintf (stderr, " Version 1.0                             \n");
   fprintf (stderr, "                                         \n");
   fprintf (stderr, " Copyright (C) 1994, Geoff Friesen B.Sc. \n");
   fprintf (stderr, " All rights reserved.                    \n");
   fprintf (stderr, "ͼ\n\n");

   if (argc != 2 && argc != 6)
   {
       fprintf (stderr, "pcx2g34: invalid number of arguments\n");
       fprintf (stderr, "syntax : pcx2g34 filespec [x y nx ny]\n");
       return 1;
   }

   if (argc == 6)
   {
       x = atoi (argv [2]);
       y = atoi (argv [3]);
       nx = atoi (argv [4]);
       ny = atoi (argv [5]);
   }

   if (!(fnsplit (argv [1], drive, dir, file, ext) & EXTENSION))
       strcpy (ext, ".PCX");

   fnmerge (path, drive, dir, file, ext);

   if ((fs = fopen (path, "rb")) == NULL)
   {
       fprintf (stderr, "pcx2g34: unable to open %s\n", path);
       return 1;
   }

   if (fread ((char *) &pcxhdr, 1, sizeof(PCXHEAD), fs) != sizeof(PCXHEAD))
   {
       fprintf (stderr, "pcx2gsi: unable to read PCX header\n");
       (void) fclose (fs);
       return 1;
   }

   if (pcxhdr.manufacturer != 0x0a)
   {
       fprintf (stderr, "pcx2gsi: %s is not a PCX file\n", path);
       (void) fclose (fs);
       return 1;
   }

   if (pcxhdr.bits_per_pixel == 1)
       bits = pcxhdr.colour_planes;
   else
       bits = pcxhdr.bits_per_pixel;

   width = (pcxhdr.xmax-pcxhdr.xmin)+1;
   depth = (pcxhdr.ymax-pcxhdr.ymin)+1;
   bytes = pcxhdr.bytes_per_line;

   fprintf (stderr, "Processing %s ...\n\n", path);

   fprintf (stderr, "Version = %u\n", pcxhdr.version);
   fprintf (stderr, "Width = %u pixels\n", width);
   fprintf (stderr, "Depth = %u pixels\n", depth);
   fprintf (stderr, "Bits per pixel = %u\n\n", bits);

   if (bits != 1 && bits != 3 && bits != 4 && bits != 8)
   {
       fprintf (stderr, "pcx2g34: unable to handle %s\n", argv [1]);
       (void) fclose (fs);
       return 1;
   }

   if (pcxhdr.bits_per_pixel == 4 && pcxhdr.colour_planes == 1)
   {
       fprintf (stderr, "pcx2g34: unable to handle this 4-bit format\n");
       (void) fclose (fs);
       return 1;
   }

   if (argc != 6)
   {
       x = 0;
       y = 0;
       nx = min (320, width);
       ny = min (400, depth);
   }

   if (nx > 320 || ny > 400 || x+nx > width || y+ny > depth)
   {
       fprintf (stderr, "pcx2g34: argument(s) out of range\n");
       (void) fclose (fs);
       return 1;
   }

   fnsplit (path, drive, dir, file, ext);
   strcpy (ext, ".G34");
   fnmerge (path, drive, dir, file, ext);

   if ((fd = fopen (path, "wb")) == NULL)
   {
       fprintf (stderr, "pcx2g34: unable to create %s\n", path);
       (void) fclose (fs);
       return 1;
   }

   memset (header, 0, 128);

   header [0] = 'G';
   header [1] = '3';
   header [2] = '4';

   *(unsigned *) (header+4) = nx;
   *(unsigned *) (header+6) = ny;

   if (fwrite (header, 1, 128, fd) != 128)
   {
	fprintf (stderr, "pcx2g34: unable to write to %s\n", path);
	(void) fclose (fs);
	(void) fclose (fd);
	(void) unlink (path);
	return 1;
   }

   _ES = FP_SEG(dac);
   _DX = FP_OFF(dac);

   asm mov ax, 1017h
   asm mov bx, 0
   asm mov cx, 256
   asm int 10h

   if (bits == 8 && pcxhdr.version >= 5)
   {
       fseek (fs, -769L, SEEK_END);
       if (fgetc (fs) == 12)
       {
	   if (fread (dac, 1, 768, fs) != 768)
	   {
	       (void) fclose (fs);
	       (void) fclose (fd);
	       fprintf (stderr, "pcx2g34: unable to read palette\n");
	       return 1;
	   }
	   else
	   for (i = 0; i < 768; i++)
		dac [i] >>= 2;
       }
       else
       {
	   memcpy (dac, pcxhdr.palette, 48);
	   for (i = 0; i < 48; i++)
		dac [i] >>= 2;
       }
       fseek (fs, (unsigned long) sizeof(PCXHEAD), SEEK_SET);
   }
   else
   if (bits != 1)
   {
       memcpy (dac, pcxhdr.palette, 48);
       for (i = 0; i < 48; i++)
	    dac [i] >>= 2;
   }

   if (fwrite (dac, 1, 768, fd) != 768)
   {
       fprintf (stderr, "pcx2g34: unable to write to %s\n", path);
       (void) fclose (fs);
       (void) fclose (fd);
       (void) unlink (path);
       return 1;
   }

   fprintf (stderr, "pcx2g34: converting %s to %s\n", argv [1], path);

   switch (bits)
   {
      case 1 : status = Copy2 (fs, fd, x, y, nx, ny, bytes, path);
	       break;

      case 3 : status = Copy8 (fs, fd, x, y, nx, ny, bytes, path);
	       break;

      case 4 : status = Copy16 (fs, fd, x, y, nx, ny, bytes, path);
	       break;

      case 8 : status = Copy256 (fs, fd, x, y, nx, ny, bytes, path);
   }

   (void) fclose (fs);
   (void) fclose (fd);

   if (!status)
       fprintf (stderr, "pcx2g34: %s successfully generated\n", path);
   else
       (void) unlink (path);

   return status;
}

int Copy2 (FILE *fs, FILE *fd, int x, int y, int nx, int ny,
	   int bytes, char *path)
{
   char *ibuffer, *obuffer;
   int bit, byte, pos, row;

   if ((ibuffer = (char *) malloc (bytes)) == NULL)
   {
       fprintf (stderr, "pcx2g34: insufficient memory\n");
       return 1;
   }

   if ((obuffer = (char *) malloc (bytes*8)) == NULL)
   {
       fprintf (stderr, "pcx2g34: insufficient memory\n");
       free (ibuffer);
       return 1;
   }

   for (row = 0; row < y; row++)
	if (ReadPcxLine (ibuffer, fs, bytes) != bytes)
	{
	    fprintf (stderr, "pcx2g34: corrupt line in PCX file\n");
	    free (ibuffer);
	    free (obuffer);
	    return 1;
	}

   for (row = 0; row < ny; row++)
   {
	if (ReadPcxLine (ibuffer, fs, bytes) != bytes)
	{
	    fprintf (stderr, "pcx2g34: corrupt line in PCX file\n");
	    free (ibuffer);
	    free (obuffer);
	    return 1;
	}

	pos = 0;
	for (byte = 0; byte < bytes; byte++)
	     for (bit = 128; bit; bit >>= 1)
		  obuffer [pos++] = (ibuffer [byte] & bit) ? 15 : 0;

	if (fwrite (obuffer+x, 1, nx, fd) != nx)
	{
	    fprintf (stderr, "pcx2g34: unable to write to %s\n", path);
	    free (ibuffer);
	    free (obuffer);
	    return 1;
	}
   }

   free (ibuffer);
   free (obuffer);
   return 0;
}

int Copy8 (FILE *fs, FILE *fd, int x, int y, int nx, int ny,
	   int bytes, char *path)
{
   char *ibuffer, *obuffer;
   int bit, byte, pos, row;

   if ((ibuffer = (char *) malloc (bytes*4)) == NULL)
   {
       fprintf (stderr, "pcx2g34: insufficient memory\n");
       return 1;
   }

   if ((obuffer = (char *) malloc (bytes*8)) == NULL)
   {
       fprintf (stderr, "pcx2g34: insufficient memory\n");
       free (ibuffer);
       return 1;
   }

   for (row = 0; row < y; row++)
	if (ReadPcxLine (ibuffer, fs, bytes*4) != bytes*4)
	{
	    fprintf (stderr, "pcx2g34: corrupt line in PCX file\n");
	    free (ibuffer);
	    free (obuffer);
	    return 1;
	}

   for (row = 0; row < ny; row++)
   {
	if (ReadPcxLine (ibuffer, fs, bytes*3) != bytes*3)
	{
	    fprintf (stderr, "pcx2g34: corrupt line in PCX file\n");
	    free (ibuffer);
	    free (obuffer);
	    return 1;
	}

	pos = 0;
	for (byte = 0; byte < bytes; byte++)
	     for (bit = 128; bit; bit >>= 1)
	     {
		  obuffer [pos] = ((ibuffer [byte] & bit) ? 1 : 0 );
		  obuffer [pos] |= ((ibuffer [byte+bytes] & bit) ? 1 : 0)
				   << 1;
		  obuffer [pos++] |= ((ibuffer [byte+2*bytes] & bit)
				     ? 1 : 0) << 2;
	     }

	if (fwrite (obuffer+x, 1, nx, fd) != nx)
	{
	    fprintf (stderr, "pcx2g34: unable to write to %s\n", path);
	    free (ibuffer);
	    free (obuffer);
	    return 1;
	}
   }

   free (ibuffer);
   free (obuffer);
   return 0;
}

int Copy16 (FILE *fs, FILE *fd, int x, int y, int nx, int ny,
	    int bytes, char *path)
{
   char *ibuffer, *obuffer;
   int bit, byte, pos, row;

   if ((ibuffer = (char *) malloc (bytes*4)) == NULL)
   {
       fprintf (stderr, "pcx2g34: insufficient memory\n");
       return 1;
   }

   if ((obuffer = (char *) malloc (bytes*8)) == NULL)
   {
       fprintf (stderr, "pcx2g34: insufficient memory\n");
       free (ibuffer);
       return 1;
   }

   for (row = 0; row < y; row++)
	if (ReadPcxLine (ibuffer, fs, bytes*4) != bytes*4)
	{
	    fprintf (stderr, "pcx2g34: corrupt line in PCX file\n");
	    free (ibuffer);
	    free (obuffer);
	    return 1;
	}

   for (row = 0; row < ny; row++)
   {
	if (ReadPcxLine (ibuffer, fs, bytes*4) != bytes*4)
	{
	    fprintf (stderr, "pcx2g34: corrupt line in PCX file\n");
	    free (ibuffer);
	    free (obuffer);
	    return 1;
	}

	pos = 0;
	for (byte = 0; byte < bytes; byte++)
	     for (bit = 128; bit; bit >>= 1)
	     {
		  obuffer [pos] = ((ibuffer [byte] & bit) ? 1 : 0);
		  obuffer [pos] |= ((ibuffer [byte+bytes] & bit) ? 1 : 0)
				   << 1;
		  obuffer [pos] |= ((ibuffer [byte+2*bytes] & bit)
				   ? 1 : 0) << 2;
		  obuffer [pos++] |= ((ibuffer [byte+3*bytes] & bit)
				     ? 1 : 0) << 3;
	     }

	if (fwrite (obuffer+x, 1, nx, fd) != nx)
	{
	    fprintf (stderr, "pcx2g34: unable to write to %s\n", path);
	    free (ibuffer);
	    free (obuffer);
	    return 1;
	}
   }

   free (ibuffer);
   free (obuffer);
   return 0;
}

int Copy256 (FILE *fs, FILE *fd, int x, int y, int nx, int ny,
	     int bytes, char *path)
{
   int row;
   char *buffer;

   if ((buffer = (char *) malloc (bytes)) == NULL)
   {
       fprintf (stderr, "pcx2g34: insufficient memory\n");
       return 1;
   }

   for (row = 0; row < y; row++)
	if (ReadPcxLine (buffer, fs, bytes) != bytes)
	{
	    fprintf (stderr, "pcx2g34: corrupt line in PCX file\n");
	    free (buffer);
	    return 1;
	}

   for (row = 0; row < ny; row++)
   {
	if (ReadPcxLine (buffer, fs, bytes) != bytes)
	{
	    fprintf (stderr, "pcx2g34: corrupt line in PCX file\n");
	    free (buffer);
	    return 1;
	}

	if (fwrite (buffer+x, 1, nx, fd) != nx)
	{
	    fprintf (stderr, "pcx2g34: unable to write to %s\n", path);
	    free (buffer);
	    return 1;
	}
   }

   free (buffer);
   return 0;
}

int ReadPcxLine (char *p, FILE *fp, unsigned bytes)
{
   int c, i, n = 0;

   memset (p, 0, bytes);

   do
   {
	c = fgetc (fp) & 255;
	if ((c & 0xc0) == 0xc0)
	{
	    i = c & 0x3f;
	    c = fgetc (fp);
	    while (i--)
	       p [n++] = c;
	}
	else
	    p [n++] = c;
   }
   while (n < bytes);

   return n;
}